package solutions.bitbadger.documents

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import kotlin.test.assertEquals

class QueryTest {

    /** Test table name */
    private val tbl = "test_table"

    /** Dummy connection string for PostgreSQL */
    private val pg = ":postgresql:"

    /** Dummy connection string for SQLite */
    private val lite = ":sqlite:"

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        Configuration.connectionString = null
    }

    @Test
    @DisplayName("statementWhere generates correctly")
    fun statementWhere() =
        assertEquals("x WHERE y", Query.statementWhere("x", "y"), "Statements not combined correctly")

    // ~~~ Where ~~~

    @Test
    @DisplayName("Where.byFields is blank when given no fields")
    fun whereByFieldsBlankIfEmpty() =
        assertEquals("", Query.Where.byFields(listOf()))

    @Test
    @DisplayName("Where.byFields generates one numeric field (PostgreSQL)")
    fun whereByFieldsOneFieldPostgres() {
        Configuration.connectionString = pg
        assertEquals("(data->>'it')::numeric = :that", Query.Where.byFields(listOf(Field.equal("it", 9, ":that"))))
    }

    @Test
    @DisplayName("Where.byFields generates one alphanumeric field (PostgreSQL)")
    fun whereByFieldsOneAlphaFieldPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'it' = :that", Query.Where.byFields(listOf(Field.equal("it", "", ":that"))))
    }

    @Test
    @DisplayName("Where.byFields generates one field (SQLite)")
    fun whereByFieldsOneFieldSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'it' = :that", Query.Where.byFields(listOf(Field.equal("it", "", ":that"))))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ default match (PostgreSQL)")
    fun whereByFieldsMultipleDefaultPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'1' = :one AND (data->>'2')::numeric = :two AND data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ default match (SQLite)")
    fun whereByFieldsMultipleDefaultSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'1' = :one AND data->>'2' = :two AND data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ ANY match (PostgreSQL)")
    fun whereByFieldsMultipleAnyPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'1' = :one OR (data->>'2')::numeric = :two OR data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                FieldMatch.ANY))
    }

    @Test
    @DisplayName("Where.byFields generates multiple fields w/ ANY match (SQLite)")
    fun whereByFieldsMultipleAnySQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'1' = :one OR data->>'2' = :two OR data->>'3' = :three",
            Query.Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                FieldMatch.ANY))
    }

    @Test
    @DisplayName("Where.byId generates defaults for alphanumeric key (PostgreSQL)")
    fun whereByIdDefaultAlphaPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'id' = :id", Query.Where.byId(docId = ""))
    }

    @Test
    @DisplayName("Where.byId generates defaults for numeric key (PostgreSQL)")
    fun whereByIdDefaultNumericPostgres() {
        Configuration.connectionString = pg
        assertEquals("(data->>'id')::numeric = :id", Query.Where.byId(docId = 5))
    }

    @Test
    @DisplayName("Where.byId generates defaults (SQLite)")
    fun whereByIdDefaultSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'id' = :id", Query.Where.byId(docId = ""))
    }

    @Test
    @DisplayName("Where.byId generates named ID (PostgreSQL)")
    fun whereByIdDefaultNamedPostgres() {
        Configuration.connectionString = pg
        assertEquals("data->>'id' = :key", Query.Where.byId<String>(":key"))
    }

    @Test
    @DisplayName("Where.byId generates named ID (SQLite)")
    fun whereByIdDefaultNamedSQLite() {
        Configuration.connectionString = lite
        assertEquals("data->>'id' = :key", Query.Where.byId<String>(":key"))
    }

    // ~~~ Definition ~~~

    @Test
    @DisplayName("Definition.ensureTableFor generates correctly")
    fun ensureTableFor() =
        assertEquals("CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)",
            Query.Definition.ensureTableFor("my.table", "JSONB"), "CREATE TABLE statement not constructed correctly")

    @Test
    @DisplayName("Definition.ensureKey generates correctly with schema")
    fun ensureKeyWithSchema() =
        assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
            Query.Definition.ensureKey("test.table", Dialect.POSTGRESQL),
            "CREATE INDEX for key statement with schema not constructed correctly")

    @Test
    @DisplayName("Definition.ensureKey generates correctly without schema")
    fun ensureKeyWithoutSchema() =
        assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_${tbl}_key ON $tbl ((data->>'id'))",
            Query.Definition.ensureKey(tbl, Dialect.SQLITE),
            "CREATE INDEX for key statement without schema not constructed correctly")

    @Test
    @DisplayName("Definition.ensureIndexOn generates multiple fields and directions")
    fun ensureIndexOnMultipleFields() =
        assertEquals(
            "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
            Query.Definition.ensureIndexOn("test.table", "gibberish", listOf("taco", "guac DESC", "salsa ASC"),
                Dialect.POSTGRESQL),
            "CREATE INDEX for multiple field statement not constructed correctly")

    @Test
    @DisplayName("Definition.ensureIndexOn generates nested PostgreSQL field")
    fun ensureIndexOnNestedPostgres() =
        assertEquals("CREATE INDEX IF NOT EXISTS idx_${tbl}_nest ON $tbl ((data#>>'{a,b,c}'))",
            Query.Definition.ensureIndexOn(tbl, "nest", listOf("a.b.c"), Dialect.POSTGRESQL),
            "CREATE INDEX for nested PostgreSQL field incorrect")

    @Test
    @DisplayName("Definition.ensureIndexOn generates nested SQLite field")
    fun ensureIndexOnNestedSQLite() =
        assertEquals("CREATE INDEX IF NOT EXISTS idx_${tbl}_nest ON $tbl ((data->'a'->'b'->>'c'))",
            Query.Definition.ensureIndexOn(tbl, "nest", listOf("a.b.c"), Dialect.SQLITE),
            "CREATE INDEX for nested SQLite field incorrect")

    @Test
    @DisplayName("insert generates correctly")
    fun insert() {
        Configuration.connectionString = pg
        assertEquals("INSERT INTO $tbl VALUES (:data)", Query.insert(tbl), "INSERT statement not constructed correctly")
    }

    @Test
    @DisplayName("save generates correctly")
    fun save() {
        Configuration.connectionString = pg
        assertEquals(
            "INSERT INTO $tbl VALUES (:data) ON CONFLICT ((data->>'id')) DO UPDATE SET data = EXCLUDED.data",
            Query.save(tbl), "INSERT ON CONFLICT UPDATE statement not constructed correctly")
    }

    @Test
    @DisplayName("count generates correctly")
    fun count() =
        assertEquals("SELECT COUNT(*) AS it FROM $tbl", Query.count(tbl), "Count query not constructed correctly")

    @Test
    @DisplayName("exists generates correctly")
    fun exists() =
        assertEquals("SELECT EXISTS (SELECT 1 FROM $tbl WHERE turkey) AS it", Query.exists(tbl, "turkey"),
            "Exists query not constructed correctly")

    @Test
    @DisplayName("find generates correctly")
    fun find() =
        assertEquals("SELECT data FROM $tbl", Query.find(tbl), "Find query not constructed correctly")

    @Test
    @DisplayName("update generates successfully")
    fun update() =
        assertEquals("UPDATE $tbl SET data = :data", Query.update(tbl), "Update query not constructed correctly")

    @Test
    @DisplayName("delete generates successfully")
    fun delete() =
        assertEquals("DELETE FROM $tbl", Query.delete(tbl), "Delete query not constructed correctly")

    @Test
    @DisplayName("orderBy generates for no fields")
    fun orderByNone() {
        assertEquals("", Query.orderBy(listOf(), Dialect.POSTGRESQL), "ORDER BY should have been blank (PostgreSQL)")
        assertEquals("", Query.orderBy(listOf(), Dialect.SQLITE), "ORDER BY should have been blank (SQLite)")
    }

    @Test
    @DisplayName("orderBy generates single, no direction for PostgreSQL")
    fun orderBySinglePostgres() =
        assertEquals(" ORDER BY data->>'TestField'",
            Query.orderBy(listOf(Field.named("TestField")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates single, no direction for SQLite")
    fun orderBySingleSQLite() =
        assertEquals(" ORDER BY data->>'TestField'", Query.orderBy(listOf(Field.named("TestField")), Dialect.SQLITE),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates multiple with direction for PostgreSQL")
    fun orderByMultiplePostgres() =
        assertEquals(" ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
            Query.orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.POSTGRESQL),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates multiple with direction for SQLite")
    fun orderByMultipleSQLite() =
        assertEquals(" ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
            Query.orderBy(
                listOf(Field.named("Nested.Test.Field DESC"), Field.named("AnotherField"), Field.named("It DESC")),
                Dialect.SQLITE),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates numeric ordering PostgreSQL")
    fun orderByNumericPostgres() =
        assertEquals(" ORDER BY (data->>'Test')::numeric",
            Query.orderBy(listOf(Field.named("n:Test")), Dialect.POSTGRESQL), "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates numeric ordering for SQLite")
    fun orderByNumericSQLite() =
        assertEquals(" ORDER BY data->>'Test'", Query.orderBy(listOf(Field.named("n:Test")), Dialect.SQLITE),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates case-insensitive ordering for PostgreSQL")
    fun orderByCIPostgres() =
        assertEquals(" ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
            Query.orderBy(listOf(Field.named("i:Test.Field DESC NULLS FIRST")), Dialect.POSTGRESQL),
            "ORDER BY not constructed correctly")

    @Test
    @DisplayName("orderBy generates case-insensitive ordering for SQLite")
    fun orderByCISQLite() =
        assertEquals(" ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
            Query.orderBy(listOf(Field.named("i:Test.Field ASC NULLS LAST")), Dialect.SQLITE),
            "ORDER BY not constructed correctly")
}
